/*
 * ********************************************************************************************************************
 *
 *                      :*$@@%$*:                         ;:                ;;    ;;
 *                    :@@%!  :!@@%:                       %!             ;%%@@%$ =@@@@@@@%;     @%@@@%%%%@@@@@
 *                   :@%;       :$=                       %%$$$%$$         ;$$  ;$@=   !@$
 *                   =@!                                  %!              @ $=;%   !@@@%:      !$$$$$$$$$$$$$$=
 *                   =@*                                  %!              @ $= % %@=   =%@!      %=
 *              *$%%! @@=        ;=$%%%$*:                %!              @ $= % =%%%%%%@$      *%:         =%
 *            %@@!:    !@@@%=$@@@@%!  :*@@$:              %!              @ $= % $*     ;@      @*          :%*
 *          ;@@!          ;!!!;:         ;@%:      =======@%========*     @ $$ % $%*****$@     :@$=*********=@$
 *          $@*   ;@@@%=!:                *@*
 *          =@$    ;;;!=%@@@@=!           =@!
 *           %@$:      =@%: :*@@@*       %@=                    Copyright (c) 2016-2019.  北京上格云技术有限公司
 *            ;%@@$=$@@%*       *@@@$=%@@%;
 *               ::;::             ::;::                                              All rights reserved.
 *
 * ********************************************************************************************************************
 */

import { SGraphyItem } from "@saga-web/graphy/";
import { SMouseEvent } from "@saga-web/base/";;
import {
    SColor,
    SLineCapStyle,
    SPainter,
    SPath2D,
    SPoint,
    SPolygonUtil,
    SRect,
    SLine
} from "@saga-web/draw";
import { PolygonData } from "./type/PolygonData";
import { Point } from "./type/Point";
import { SRelationState } from "@saga-web/big/lib/enums/SRelationState"
import { SMathUtil } from "@saga-web/big/lib/utils/SMathUtil";

/**
 * 编辑多边形
 *
 * @author  韩耀龙
 */
export class SPolygonItem extends SGraphyItem {
    /** X坐标最小值  */
    private minX = Number.MAX_SAFE_INTEGER;
    /** X坐标最大值  */
    private maxX = Number.MIN_SAFE_INTEGER;
    /** Y坐标最小值  */
    private minY = Number.MAX_SAFE_INTEGER;
    /** Y坐标最大值  */
    private maxY = Number.MIN_SAFE_INTEGER;
    /** 轮廓线坐标  */
    pointList: SPoint[] = [];
    /** 是否闭合    */
    closeFlag: boolean = false;
    // 当前状态 1：show 2 创建中，3 编辑中
    _status: number = SRelationState.Normal;
    // 获取当前状态
    get getStatus(): number {
        return this._status;
    }
    // 编辑当前状态
    set setStatus(value: number) {
        this._status = value;
        this.update();
    };
    data: PolygonData | null = null;
    /** 边框颜色 */
    borderColor: SColor = new SColor("#0091FF");
    /** 填充颜色 */
    brushColor: SColor = new SColor("#1EE887");
    /** border宽 只可输入像素宽*/
    borderLine: number = 4;
    /** 鼠标移动点  */
    private lastPoint = new SPoint();
    /** 当前鼠标获取顶点对应索引 */
    private curIndex: number = -1;
    /** 灵敏像素 */
    private len: number = 15;
    /** 场景像素  */
    private scenceLen: number = 15;
    /**
    * 构造函数
    *
    * @param parent    指向父对象
    * @param data      PolygonData数据
    */

    constructor(parent: SGraphyItem | null, data: PolygonData | null) {
        super(parent);
        this.data = data;

        if (data && data.PointList) {
            this._xyzListToSPointList(data.PointList);
        };
        this.zOrder = 100;
    }


    /**
    * 储存新的多边形顶点
    * @param x   点位得x坐标
    * @param y   点位得y坐标
    * @param i   储存所在索引
    * @return SPoint
    */

    reservePoint(x: number, y: number, i?: number): SPoint {
        const point = new SPoint(x, y);
        if (typeof i == 'undefined') {
            this.pointList.push(point)
        } else {
            this.pointList.splice(i, 0, point);
        }
        this.update()
        return point
    }


    /**
    * 删除点位
    * @param i   删除点所在的索引
    * @return SPoint
    */

    deletePoint(i?: number): SPoint | null | undefined {
        let point = null;
        if (typeof i != 'undefined') {
            if (this.pointList.length - 1 >= i) {
                point = this.pointList[i];
                this.pointList.splice(i, 0);
            } else {
                point = null
            }
        } else {
            point = this.pointList.pop();
        }
        return point
    }


    /**
    * 多边形顶点的移动位置
    * @param x   点位得x坐标
    * @param y   点位得y坐标
    * @param i   点位得i坐标
    * @return SPoint
    */

    movePoint(x: number, y: number, i: number, ): SPoint | null | undefined {
        let point = null;
        if (this.pointList.length) {
            this.pointList[i].x = x;
            this.pointList[i].y = y;
        }
        point = this.pointList[i];
        return point
    }


    /**
    * 打印出多边形数组
    * @return SPoint[]
    */

    PrintPointList(): SPoint[] {
        return this.pointList
    }

    /**
     * Item对象边界区域
     *
     * @return SRect
     */
    boundingRect(): SRect {
        return new SRect(
            this.minX,
            this.minY,
            this.maxX - this.minX,
            this.maxY - this.minY
        );
    } // Function boundingRect()


    /**
    * xyz数据转换为SPoint格式函数；获取外接矩阵参数
    *
    * @param arr    外层传入的数据PointList
    */

    protected _xyzListToSPointList(arr: Point[]) {
        if (arr.length) {
            this.pointList = arr.map(it => {
                let x = it.X,
                    y = it.Y;
                if (x < this.minX) {
                    this.minX = x;
                }
                if (y < this.minY) {
                    this.minY = y;
                }
                if (x > this.maxX) {
                    this.maxX = x;
                }
                if (y > this.maxY) {
                    this.maxY = y;
                }
                return new SPoint(x, y);
            });
        } else {
            this.pointList = []
        }
    }


    /**
    * 构造函数  展示多边形轮廓
    *
    */

    protected showPolygon(painter: SPainter, pointList: SPoint[]) {
        painter.pen.lineCapStyle = SLineCapStyle.Square;
        painter.pen.color = this.borderColor;
        painter.pen.lineWidth = painter.toPx(this.borderLine);
        painter.brush.color = this.brushColor;
        painter.drawPolygon([...this.pointList]);
    }


    /**
    * 构造函数  创建多边形
    *
    */

    protected createPolygon(painter: SPainter, pointList: SPoint[]) {
        painter.pen.lineCapStyle = SLineCapStyle.Square;
        painter.pen.color = new SColor("#0091FF");
        painter.pen.lineWidth = painter.toPx(4);
        painter.drawPolyline(this.pointList);
        if (this.lastPoint) {
            painter.drawLine(
                pointList[this.pointList.length - 1],
                this.lastPoint
            );
        };
        painter.pen.color = SColor.Transparent;
        painter.pen.lineWidth = painter.toPx(4);
        painter.drawPolygon([...this.pointList, this.lastPoint]);
    }


    /**
    * 构造函数  创建多边形
    *
    */

    protected editPolygon(painter: SPainter, pointList: SPoint[]) {
        // 展示多边形
        this.showPolygon(painter, pointList);
        // 绘制顶点块
        pointList.forEach((item, index) => {
            painter.drawCircle(item.x, item.y, painter.toPx(8))
        })
    }




    ////////////以下为鼠标事件

    /**
    * 鼠标双击事件
    *
    * @param	event         事件参数
    * @return	boolean
    */

    onDoubleClick(event: SMouseEvent): boolean {
        // 如果位show状态 双击改对象则需改为编辑状态
        if (this._status == 1) {
            this._status = 3;
            this.update()
        }
        return true;
    } // Function onDoubleClick()


    /**
    * 键盘事件
    *
    * @param	event         事件参数
    * @return	boolean
    */
    onKeyDown(event: KeyboardEvent): boolean {
        if (this._status == 1) {
            return false;
        } else if (this._status == 2) {
            if (event.code == 'Enter') {
                this._status = 1
            }
        } else if (this._status == 3) {
            // if (event.code == 'Enter') {
            //     this._status = 1
            // }
        } else {

        }
        this.update()
        return true;
    } // Function onKeyDown()
    /**
     * 鼠标双击事件
     *
     * @param	event         事件参数
     * @return	boolean
     */
    onMouseDown(event: SMouseEvent): boolean {
        // 如果状态为编辑状态则添加点
        if (this._status == 2) {
            // 新增顶点
            this.reservePoint(event.x, event.y)
        } else if (this._status == 3) {
            // 1 判断是否点击在多边形顶点
            // 敏感度
            this.len = 15;
            this.pointList.forEach((item, index) => {
                let dis = SMathUtil.pointDistance(
                    event.x,
                    event.y,
                    item.x,
                    item.y
                );
                if (dis < this.len) {
                    this.len = dis;
                    this.curIndex = index;
                }
            });
            // 判断是否点击在多边形得边上
            if (-1 == this.curIndex) {
                let len = SMathUtil.pointToLine(
                    new SPoint(event.x, event.y),
                    new SLine(this.pointList[0], this.pointList[1])
                ),
                    index = 0;
                if (this.pointList.length > 2) {
                    for (let i = 1; i < this.pointList.length; i++) {
                        let dis = SMathUtil.pointToLine(
                            new SPoint(event.x, event.y),
                            new SLine(this.pointList[i], this.pointList[i + 1])
                        );
                        if (i == this.pointList.length) {
                            console.log(this.pointList.length)
                            dis = SMathUtil.pointToLine(
                                new SPoint(event.x, event.y),
                                new SLine(this.pointList[i], this.pointList[0])
                            );
                        }
                        if (dis.MinDis < len.MinDis) {
                            len = dis;
                            index = i;
                        }
                    }
                }
                if (len.Point) {
                    if (len.MinDis <= this.scenceLen) {
                        this.pointList.splice(index + 1, 0, len.Point);
                        this.update();
                    }
                }
            }
        }
        return true;
    } // Function onMouseDown()


    /**
     * 鼠标双击事件
     *
     * @param	event         事件参数
     * @return	boolean
     */
    onMouseEnter(event: SMouseEvent): boolean {
        return true;
    } // Function onMouseDown()
    /**
     * 鼠标双击事件
     *
     * @param	event         事件参数
     * @return	boolean
     */
    onMouseLeave(event: SMouseEvent): boolean {
        return true;
    } // Function onMouseDown()
    onMouseMove(event: SMouseEvent): boolean {
        if (this._status == 2) {
            this.lastPoint.x = event.x;
            this.lastPoint.y = event.y;
        } else if (this._status == 3) {
            if (-1 != this.curIndex) {
                console.log('onMouseMove')
                this.pointList[this.curIndex].x = event.x;
                this.pointList[this.curIndex].y = event.y;
            }
        }
        this.update()
        return true;
    } // Function onMouseDown()
    onMouseUp(event: SMouseEvent): boolean {
        console.log('onMouseUp')
        if (this._status == 3) {
            this.curIndex = -1;
        }
        return true;
    } // Function onMouseUp()
    onResize(event: SMouseEvent): boolean {
        return true;
    } // Function onMouseUp()



    /**
    *  点击点是否在item范围内
    *
    */
    // contains(x: number, y: number): boolean {
    //     console.log('PTL.MinDis < this.scenceLen')
    //     if (3 == this._status) {
    //         let p = new SPoint(x, y);
    //         //  todo差缩放
    //         for (let i = 1; i < this.pointList.length; i++) {
    //             let PTL = SMathUtil.pointToLine(
    //                 p,
    //                 new SLine(
    //                     this.pointList[i - 1].x,
    //                     this.pointList[i - 1].y,
    //                     this.pointList[i].x,
    //                     this.pointList[i].y
    //                 )
    //             );
    //             if (PTL.MinDis < this.scenceLen) {
    //                 console.log('PTL.MinDis < this.scenceLen', PTL.MinDis, this.scenceLen)
    //                 return true;
    //             }
    //         }
    //         return false;
    //     }
    //     return false;
    // }

    /**
    * Item绘制操作
    *
    * @param   painter       painter对象
    */

    onDraw(painter: SPainter): void {
        this.scenceLen = painter.toPx(this.len)
        // 当状态为展示状态
        if (this._status == 1) {
            // 闭合多边形
            this.showPolygon(painter, this.pointList);
        } else if (this._status == 2) {
            // 创建状态
            this.createPolygon(painter, this.pointList)
        } else {
            // 编辑状态
            this.editPolygon(painter, this.pointList);
        }
    } // Function onDraw()

}
